from __future__ import annotations

import json
import os
import sys
from dataclasses import dataclass
from dataclasses import field as dataclass_field
from functools import lru_cache
from logging import getLogger
from pathlib import Path
from typing import Any, Callable, get_args
from typing import Union as TypingUnion

import torch
from arpeggio import NoMatch
from arpeggio.cleanpeg import ParserPEG
from sympy import *
from sympy import core, srepr

from qadence import QNN, QuantumCircuit, QuantumModel, operations, parameters
from qadence import blocks as qadenceblocks
from qadence.blocks import AbstractBlock
from qadence.blocks.utils import tag
from qadence.parameters import Parameter
from qadence.register import Register
from qadence.types import SerializationFormat

# Modules to be automatically added to the qadence namespace
__all__ = ["deserialize", "load", "save", "serialize"]


logger = getLogger(__name__)


def file_extension(file: Path | str) -> str:
    FORMAT = ""
    if isinstance(file, str):
        _, extension = os.path.splitext(file)
        FORMAT = extension[1:].upper()
    elif isinstance(file, os.PathLike):
        _, extension = os.path.splitext(str(file))
        FORMAT = extension[1:].upper()
    return FORMAT


SUPPORTED_OBJECTS = [
    AbstractBlock,
    QuantumCircuit,
    QuantumModel,
    QNN,
    Register,
    core.Basic,
    torch.nn.Module,
]
SUPPORTED_TYPES = TypingUnion[
    AbstractBlock,
    QuantumCircuit,
    QuantumModel,
    Register,
    core.Basic,
    torch.nn.Module,
]

ALL_BLOCK_NAMES = [
    n for n in dir(qadenceblocks) if not (n.startswith("__") and n.endswith("__"))
] + [n for n in dir(operations) if not (n.startswith("__") and n.endswith("__"))]
SYMPY_EXPRS = [n for n in dir(core) if not (n.startswith("__") and n.endswith("__"))]
QADENCE_PARAMETERS = [n for n in dir(parameters) if not (n.startswith("__") and n.endswith("__"))]


THIS_PATH = Path(__file__).parent
GRAMMAR_FILE = THIS_PATH / "serial_expr_grammar.peg"


@lru_cache
def _parser_fn() -> ParserPEG:
    with open(GRAMMAR_FILE, "r") as f:
        grammar = f.read()
    return ParserPEG(grammar, "Program")


_parsing_serialize_expr = _parser_fn()


def parse_expr_fn(code: str) -> bool:
    """
    A parsing expressions function that checks whether a given code is valid on.

    the parsing grammar. The grammar is defined to be compatible with `sympy`
    expressions, such as `Float('-0.33261030434342942', precision=53)`, while
    avoiding code injection such as `2*3` or `__import__('os').system('ls -la')`.

    Args:
        code (str): code to be parsed and checked.

    Returns:
        Boolean indicating whether the code matches the defined grammar or not.
    """

    parser = _parsing_serialize_expr
    try:
        parser.parse(code)
    except NoMatch:
        return False
    else:
        return True


@dataclass
class SerializationModel:
    """
    A serialization model class to serialize data from `QuantumModel`s,.

    `torch.nn.Module` and similar structures. The data included in the
    serialization logic includes: the `AbstractBlock` and its children
    classes, `QuantumCircuit`, `Register`, and `sympy` expressions
    (including `Parameter` class from `qadence.parameters`).

    A children class must define the `value` attribute type and how to
    handle it, since it is the main property for the class to be used
    by the serialization process. For instance:

    ```python
    @dataclass
    class QuantumCircuitSerialization(SerializationModel):
        value: QuantumCircuit = dataclass_field(init=False)

        def __post_init__(self) -> None:
            self.value = (
                QuantumCircuit._from_dict(self.d)
                if isinstance(self.d, dict)
                else self.d
            )
    ```
    """

    d: dict = dataclass_field(default_factory=dict)
    value: Any = dataclass_field(init=False)


@dataclass
class BlockTypeSerialization(SerializationModel):
    value: AbstractBlock = dataclass_field(init=False)

    def __post_init__(self) -> None:
        block = (
            getattr(operations, self.d["type"])
            if hasattr(operations, self.d["type"])
            else getattr(qadenceblocks, self.d["type"])
        )._from_dict(self.d)
        if self.d["tag"] is not None:
            block = tag(block, self.d["tag"])
        self.value = block


@dataclass
class QuantumCircuitSerialization(SerializationModel):
    value: QuantumCircuit = dataclass_field(init=False)

    def __post_init__(self) -> None:
        self.value = QuantumCircuit._from_dict(self.d) if isinstance(self.d, dict) else self.d


@dataclass
class RegisterSerialization(SerializationModel):
    value: Register = dataclass_field(init=False)

    def __post_init__(self) -> None:
        self.value = Register._from_dict(self.d)


@dataclass
class ModelSerialization(SerializationModel):
    as_torch: bool = False
    value: torch.nn.Module = dataclass_field(init=False)

    def __post_init__(self) -> None:
        module_name = list(self.d.keys())[0]
        obj = globals().get(module_name, None)
        if obj is None:
            obj = self._resolve_module(module_name)
        if hasattr(obj, "_from_dict"):
            self.value = obj._from_dict(self.d, self.as_torch)
        elif hasattr(obj, "load_state_dict"):
            self.value = obj.load_state_dict(self.d[module_name])
        else:
            msg = (
                f"Unable to deserialize object '{module_name}'. "
                f"Supported types are {SUPPORTED_OBJECTS}."
            )
            logger.error(TypeError(msg))
            raise TypeError(msg)

    @staticmethod
    def _resolve_module(module: str) -> Any:
        for loaded_module in sys.modules.keys():
            if "qadence" in loaded_module:
                obj = getattr(sys.modules[loaded_module], module, None)
                if obj:
                    return obj
        raise ValueError(f"Couldn't resolve module '{module}'.")


@dataclass
class ExpressionSerialization(SerializationModel):
    value: str | core.Expr | float = dataclass_field(init=False)

    def __post_init__(self) -> None:
        if parse_expr_fn(self.d["expression"]):
            expr = eval(self.d["expression"])
            if hasattr(expr, "free_symbols"):
                for s in expr.free_symbols:
                    s.value = float(self.d["symbols"][s.name]["value"])
            self.value = expr
        else:
            raise ValueError(f"Invalid expression: {self.d['expression']}")


def save_pt(d: dict, file_path: str | Path) -> None:
    torch.save(d, file_path)


def save_json(d: dict, file_path: str | Path) -> None:
    with open(file_path, "w") as file:
        file.write(json.dumps(d))


def load_pt(file_path: str | Path, map_location: str) -> Any:
    return torch.load(file_path, map_location=map_location)


def load_json(file_path: str | Path, map_location: str) -> Any:
    with open(file_path, "r") as file:
        return json.load(file)


FORMAT_DICT = {
    SerializationFormat.PT: (".pt", save_pt, load_pt, True),
    SerializationFormat.JSON: (".json", save_json, load_json, False),
}


def serialize(obj: SUPPORTED_TYPES, save_params: bool = False) -> dict:
    """
    Supported Types:

    AbstractBlock | QuantumCircuit | QuantumModel | torch.nn.Module | Register | Module
    Serializes a qadence object to a dictionary.

    Arguments:
        obj (AbstractBlock | QuantumCircuit | QuantumModel | Register | torch.nn.Module):
    Returns:
        A dict.

    Examples:
    ```python exec="on" source="material-block" result="json"
    import torch
    from qadence import serialize, deserialize, hea, hamiltonian_factory, Z
    from qadence import QuantumCircuit, QuantumModel

    n_qubits = 2
    myblock = hea(n_qubits=n_qubits, depth=1)
    block_dict = serialize(myblock)
    print(block_dict)

    ## Lets use myblock in a QuantumCircuit and serialize it.

    qc = QuantumCircuit(n_qubits, myblock)
    qc_dict = serialize(qc)
    qc_deserialized = deserialize(qc_dict)
    assert qc == qc_deserialized

    ## Finally, let's wrap it in a QuantumModel
    obs = hamiltonian_factory(n_qubits, detuning = Z)
    qm = QuantumModel(qc, obs, backend='pyqtorch', diff_mode='ad')

    qm_dict = serialize(qm)
    qm_deserialized = deserialize(qm_dict)
    # Lets check if the loaded QuantumModel returns the same expectation
    assert torch.isclose(qm.expectation({}), qm_deserialized.expectation({}))
    ```
    """
    if not isinstance(obj, get_args(SUPPORTED_TYPES)):
        logger.error(TypeError(f"Serialization of object type {type(obj)} not supported."))

    d: dict = dict()
    try:
        if isinstance(obj, core.Expr):
            symb_dict = dict()
            expr_dict = {"name": str(obj), "expression": srepr(obj)}
            symbs: set[Parameter | core.Basic] = obj.free_symbols
            if symbs:
                symb_dict = {"symbols": {str(s): s._to_dict() for s in symbs}}
            d = {**expr_dict, **symb_dict}
        else:
            if hasattr(obj, "_to_dict"):
                model_to_dict: Callable = obj._to_dict
                d = (
                    model_to_dict(save_params)
                    if isinstance(obj, torch.nn.Module)
                    else model_to_dict()
                )
            elif hasattr(obj, "state_dict"):
                d = {type(obj).__name__: obj.state_dict()}
            else:
                raise ValueError(f"Cannot serialize object {obj}.")
    except Exception as e:
        logger.error(f"Serialization of object {obj} failed due to {e}")
    return d


def deserialize(d: dict, as_torch: bool = False) -> SUPPORTED_TYPES:
    """
    Supported Types:

    AbstractBlock | QuantumCircuit | QuantumModel | Register | torch.nn.Module
    Deserializes a dict to one of the supported types.

    Arguments:
        d (dict): A dict containing a serialized object.
        as_torch (bool): Whether to transform to torch for the deserialized object.
    Returns:
        AbstractBlock, QuantumCircuit, QuantumModel, Register, torch.nn.Module.

    Examples:
    ```python exec="on" source="material-block" result="json"
    import torch
    from qadence import serialize, deserialize, hea, hamiltonian_factory, Z
    from qadence import QuantumCircuit, QuantumModel

    n_qubits = 2
    myblock = hea(n_qubits=n_qubits, depth=1)
    block_dict = serialize(myblock)
    print(block_dict)

    ## Lets use myblock in a QuantumCircuit and serialize it.

    qc = QuantumCircuit(n_qubits, myblock)
    qc_dict = serialize(qc)
    qc_deserialized = deserialize(qc_dict)
    assert qc == qc_deserialized

    ## Finally, let's wrap it in a QuantumModel
    obs = hamiltonian_factory(n_qubits, detuning = Z)
    qm = QuantumModel(qc, obs, backend='pyqtorch', diff_mode='ad')

    qm_dict = serialize(qm)
    qm_deserialized = deserialize(qm_dict)
    # Lets check if the loaded QuantumModel returns the same expectation
    assert torch.isclose(qm.expectation({}), qm_deserialized.expectation({}))
    ```
    """
    obj: SerializationModel
    if d.get("expression"):
        obj = ExpressionSerialization(d)
    elif d.get("block") and d.get("register"):
        obj = QuantumCircuitSerialization(d)
    elif d.get("graph"):
        obj = RegisterSerialization(d)
    elif d.get("type"):
        obj = BlockTypeSerialization(d)
    else:
        obj = ModelSerialization(d, as_torch=as_torch)
    return obj.value


def save(
    obj: SUPPORTED_TYPES,
    folder: str | Path,
    file_name: str = "",
    format: SerializationFormat = SerializationFormat.JSON,
) -> None:
    """
    Same as serialize/deserialize but for storing/loading files.

    Supported types:
    AbstractBlock | QuantumCircuit | QuantumModel | Register | torch.nn.Module
    Saves a qadence object to a json/.pt.

    Arguments:
        obj (AbstractBlock | QuantumCircuit | QuantumModel | Register):
                Either AbstractBlock, QuantumCircuit, QuantumModel, Register.
        file_name (str): The name of the file.
        format (str): The type of file to save.
    Returns:
        None.

    Examples:
    ```python exec="on" source="material-block" result="json"
    import torch
    from pathlib import Path
    import os

    from qadence import save, load, hea, hamiltonian_factory, Z
    from qadence import QuantumCircuit, QuantumModel

    n_qubits = 2
    myblock = hea(n_qubits=n_qubits, depth=1)
    qc = QuantumCircuit(n_qubits, myblock)
    # Lets store the circuit in a json file
    save(qc, '.', 'circ')
    loaded_qc = load(Path('circ.json'))
    qc == loaded_qc
    os.remove('circ.json')
    ## Let's wrap it in a QuantumModel and store that
    obs = hamiltonian_factory(n_qubits, detuning = Z)
    qm = QuantumModel(qc, obs, backend='pyqtorch', diff_mode='ad')
    save(qm, folder= '.',file_name= 'quantum_model')
    qm_loaded = load('quantum_model.json')
    os.remove('quantum_model.json')
    ```
    """
    if not isinstance(obj, get_args(SUPPORTED_TYPES)):
        logger.error(f"Serialization of object type {type(obj)} not supported.")
    folder = Path(folder)
    if not folder.is_dir():
        logger.error(NotADirectoryError)
    if file_name == "":
        file_name = type(obj).__name__
    try:
        suffix, save_fn, _, save_params = FORMAT_DICT[format]
        d = serialize(obj, save_params)
        file_path = folder / Path(file_name + suffix)
        save_fn(d, file_path)
        logger.debug(f"Successfully saved {obj} from to {folder}.")
    except Exception as e:
        logger.error(f"Unable to write {type(obj)} to disk due to {e}")


def load(file_path: str | Path, map_location: str = "cpu") -> SUPPORTED_TYPES:
    """
    Same as serialize/deserialize but for storing/loading files.

    Supported types: AbstractBlock | QuantumCircuit | QuantumModel | Register
    Loads a .json or .pt file to one of the supported types.

    Arguments:
        file_path (str): The name of the file.
        map_location (str): In case of a .pt file, on which device to load the object (cpu,cuda).
    Returns:
        A object of type AbstractBlock, QuantumCircuit, QuantumModel, Register.

    Examples:
    ```python exec="on" source="material-block" result="json"
    import torch
    from pathlib import Path
    import os

    from qadence import save, load, hea, hamiltonian_factory, Z
    from qadence import QuantumCircuit, QuantumModel

    n_qubits = 2
    myblock = hea(n_qubits=n_qubits, depth=1)
    qc = QuantumCircuit(n_qubits, myblock)
    # Lets store the circuit in a json file
    save(qc, '.', 'circ')
    loaded_qc = load(Path('circ.json'))
    qc == loaded_qc
    os.remove('circ.json')
    ## Let's wrap it in a QuantumModel and store that
    obs = hamiltonian_factory(n_qubits, detuning = Z)
    qm = QuantumModel(qc, obs, backend='pyqtorch', diff_mode='ad')
    save(qm, folder= '.',file_name= 'quantum_model')
    qm_loaded = load('quantum_model.json')
    os.remove('quantum_model.json')
    ```
    """
    d = {}
    if isinstance(file_path, str):
        file_path = Path(file_path)
    if not os.path.exists(file_path):
        logger.error(f"File {file_path} not found.")
        raise FileNotFoundError
    FORMAT = file_extension(file_path)
    _, _, load_fn, _ = FORMAT_DICT[FORMAT]  # type: ignore[index]
    try:
        d = load_fn(file_path, map_location)
        logger.debug(f"Successfully loaded {d} from {file_path}.")
    except Exception as e:
        logger.error(f"Unable to load Object from {file_path} due to {e}")
    return deserialize(d)



from __future__ import annotations

import importlib
from logging import getLogger
from typing import Callable

import nevergrad as ng
from nevergrad.optimization.base import Optimizer as NGOptimizer
from rich.progress import (
    BarColumn,
    Progress,
    TaskProgressColumn,
    TextColumn,
    TimeRemainingColumn,
)
from torch import Tensor
from torch.nn import Module
from torch.utils.data import DataLoader
from torch.utils.tensorboard import SummaryWriter

from qadence.ml_tools.config import TrainConfig
from qadence.ml_tools.data import DictDataLoader
from qadence.ml_tools.parameters import get_parameters, set_parameters
from qadence.ml_tools.printing import (
    log_model_tracker,
    log_tracker,
    plot_tracker,
    print_metrics,
    write_tracker,
)
from qadence.ml_tools.saveload import load_checkpoint, write_checkpoint
from qadence.ml_tools.tensors import promote_to_tensor
from qadence.types import ExperimentTrackingTool

logger = getLogger(__name__)


def train(
    model: Module,
    dataloader: DictDataLoader | DataLoader | None,
    optimizer: NGOptimizer,
    config: TrainConfig,
    loss_fn: Callable,
) -> tuple[Module, NGOptimizer]:
    """Runs the training loop with a gradient-free optimizer.

    Assumes that `loss_fn` returns a tuple of (loss, metrics: dict), where
    `metrics` is a dict of scalars. Loss and metrics are written to
    tensorboard. Checkpoints are written every `config.checkpoint_every` steps
    (and after the last training step).  If a checkpoint is found at `config.folder`
    we resume training from there.  The tensorboard logs can be viewed via
    `tensorboard --logdir /path/to/folder`.

    Args:
        model: The model to train
        dataloader: Dataloader constructed via `dictdataloader`
        optimizer: The optimizer to use taken from the Nevergrad library. If this is not
            the case the function will raise an AssertionError
        config: `TrainConfig` with additional training options.
        loss_fn: Loss function returning (loss: float, metrics: dict[str, float])
    """
    init_iter = 0
    if config.folder:
        model, optimizer, init_iter = load_checkpoint(config.folder, model, optimizer)
        logger.debug(f"Loaded model and optimizer from {config.folder}")

    def _update_parameters(
        data: Tensor | None, ng_params: ng.p.Array
    ) -> tuple[float, dict, ng.p.Array]:
        loss, metrics = loss_fn(model, data)  # type: ignore[misc]
        optimizer.tell(ng_params, float(loss))
        ng_params = optimizer.ask()  # type: ignore [assignment]
        params = promote_to_tensor(ng_params.value, requires_grad=False)
        set_parameters(model, params)
        return loss, metrics, ng_params

    assert loss_fn is not None, "Provide a valid loss function"
    # TODO: support also Scipy optimizers
    assert isinstance(optimizer, NGOptimizer), "Use only optimizers from the Nevergrad library"

    # initialize tracking tool
    if config.tracking_tool == ExperimentTrackingTool.TENSORBOARD:
        writer = SummaryWriter(config.folder, purge_step=init_iter)
    else:
        writer = importlib.import_module("mlflow")

    # set optimizer configuration and initial parameters
    optimizer.budget = config.max_iter
    optimizer.enable_pickling()

    # TODO: Make it GPU compatible if possible
    params = get_parameters(model).detach().numpy()
    ng_params = ng.p.Array(init=params)

    # serial training
    # TODO: Add a parallelization using the num_workers argument in Nevergrad
    progress = Progress(
        TextColumn("[progress.description]{task.description}"),
        BarColumn(),
        TaskProgressColumn(),
        TimeRemainingColumn(elapsed_when_finished=True),
    )
    with progress:
        dl_iter = iter(dataloader) if dataloader is not None else None

        for iteration in progress.track(range(init_iter, init_iter + config.max_iter)):
            if dataloader is None:
                loss, metrics, ng_params = _update_parameters(None, ng_params)

            elif isinstance(dataloader, (DictDataLoader, DataLoader)):
                data = next(dl_iter)  # type: ignore[arg-type]
                loss, metrics, ng_params = _update_parameters(data, ng_params)

            else:
                raise NotImplementedError("Unsupported dataloader type!")

            if iteration % config.print_every == 0 and config.verbose:
                print_metrics(loss, metrics, iteration)

            if iteration % config.write_every == 0:
                write_tracker(writer, loss, metrics, iteration, tracking_tool=config.tracking_tool)

            if iteration % config.plot_every == 0:
                plot_tracker(
                    writer,
                    model,
                    iteration,
                    config.plotting_functions,
                    tracking_tool=config.tracking_tool,
                )

            if config.folder:
                if iteration % config.checkpoint_every == 0:
                    write_checkpoint(config.folder, model, optimizer, iteration)

            if iteration >= init_iter + config.max_iter:
                break

    # writing hyperparameters
    if config.hyperparams:
        log_tracker(writer, config.hyperparams, metrics, tracking_tool=config.tracking_tool)

    if config.log_model:
        log_model_tracker(writer, model, dataloader, tracking_tool=config.tracking_tool)

    # Final writing and checkpointing
    if config.folder:
        write_checkpoint(config.folder, model, optimizer, iteration)
    write_tracker(writer, loss, metrics, iteration, tracking_tool=config.tracking_tool)

    # close tracker
    if config.tracking_tool == ExperimentTrackingTool.TENSORBOARD:
        writer.close()
    elif config.tracking_tool == ExperimentTrackingTool.MLFLOW:
        writer.end_run()

    return model, optimizer



from __future__ import annotations

import datetime
import os
from dataclasses import dataclass, field, fields
from logging import getLogger
from pathlib import Path
from typing import Callable, Type
from uuid import uuid4

from sympy import Basic
from torch import Tensor

from qadence.blocks.analog import AnalogBlock
from qadence.blocks.primitive import ParametricBlock
from qadence.operations import RX, AnalogRX
from qadence.parameters import Parameter
from qadence.types import (
    AnsatzType,
    BasisSet,
    ExperimentTrackingTool,
    LoggablePlotFunction,
    MultivariateStrategy,
    ReuploadScaling,
    Strategy,
)

logger = getLogger(__file__)


@dataclass
class TrainConfig:
    """Default config for the train function.

    The default value of
    each field can be customized with the constructor:

    ```python exec="on" source="material-block" result="json"
    from qadence.ml_tools import TrainConfig
    c = TrainConfig(folder="/tmp/train")
    print(str(c)) # markdown-exec: hide
    ```
    """

    max_iter: int = 10000
    """Number of training iterations."""
    print_every: int = 1000
    """Print loss/metrics."""
    write_every: int = 50
    """Write loss and metrics with the tracking tool."""
    checkpoint_every: int = 5000
    """Write model/optimizer checkpoint."""
    plot_every: int = 5000
    """Write figures."""
    log_model: bool = False
    """Logs a serialised version of the model."""
    folder: Path | None = None
    """Checkpoint/tensorboard logs folder."""
    create_subfolder_per_run: bool = False
    """Checkpoint/tensorboard logs stored in subfolder with name `<timestamp>_<PID>`.

    Prevents continuing from previous checkpoint, useful for fast prototyping.
    """
    checkpoint_best_only: bool = False
    """Write model/optimizer checkpoint only if a metric has improved."""
    val_every: int | None = None
    """Calculate validation metric.

    If None, validation check is not performed.
    """
    val_epsilon: float = 1e-5
    """Safety margin to check if validation loss is smaller than the lowest.

    validation loss across previous iterations.
    """
    validation_criterion: Callable | None = None
    """A boolean function which evaluates a given validation metric is satisfied."""
    trainstop_criterion: Callable | None = None
    """A boolean function which evaluates a given training stopping metric is satisfied."""
    batch_size: int = 1
    """The batch_size to use when passing a list/tuple of torch.Tensors."""
    verbose: bool = True
    """Whether or not to print out metrics values during training."""
    tracking_tool: ExperimentTrackingTool = ExperimentTrackingTool.TENSORBOARD
    """The tracking tool of choice."""
    hyperparams: dict = field(default_factory=dict)
    """Hyperparameters to track."""
    plotting_functions: tuple[LoggablePlotFunction, ...] = field(default_factory=tuple)  # type: ignore
    """Functions for in-train plotting."""

    # tensorboard only allows for certain types as hyperparameters
    _tb_allowed_hyperparams_types: tuple = field(
        default=(int, float, str, bool, Tensor), init=False, repr=False
    )

    def _filter_tb_hyperparams(self) -> None:
        keys_to_remove = [
            key
            for key, value in self.hyperparams.items()
            if not isinstance(value, TrainConfig._tb_allowed_hyperparams_types)
        ]
        if keys_to_remove:
            logger.warning(
                f"Tensorboard cannot log the following hyperparameters: {keys_to_remove}."
            )
            for key in keys_to_remove:
                self.hyperparams.pop(key)

    def __post_init__(self) -> None:
        if self.folder:
            if isinstance(self.folder, str):  # type: ignore [unreachable]
                self.folder = Path(self.folder)  # type: ignore [unreachable]
            if self.create_subfolder_per_run:
                subfoldername = (
                    datetime.datetime.now().strftime("%Y%m%dT%H%M%S") + "_" + hex(os.getpid())[2:]
                )
                self.folder = self.folder / subfoldername
        if self.trainstop_criterion is None:
            self.trainstop_criterion = lambda x: x <= self.max_iter
        if self.validation_criterion is None:
            self.validation_criterion = lambda *x: False
        if self.hyperparams and self.tracking_tool == ExperimentTrackingTool.TENSORBOARD:
            self._filter_tb_hyperparams()
        if self.tracking_tool == ExperimentTrackingTool.MLFLOW:
            self._mlflow_config = MLFlowConfig()
        if self.plotting_functions and self.tracking_tool != ExperimentTrackingTool.MLFLOW:
            logger.warning("In-training plots are only available with mlflow tracking.")
        if not self.plotting_functions and self.tracking_tool == ExperimentTrackingTool.MLFLOW:
            logger.warning("Tracking with mlflow, but no plotting functions provided.")

    @property
    def mlflow_config(self) -> MLFlowConfig:
        if self.tracking_tool == ExperimentTrackingTool.MLFLOW:
            return self._mlflow_config
        else:
            raise AttributeError(
                "mlflow_config is available only for with the mlflow tracking tool."
            )


class MLFlowConfig:
    """
    Configuration for mlflow tracking.

    Example:

        export MLFLOW_TRACKING_URI=tracking_uri
        export MLFLOW_EXPERIMENT=experiment_name
        export MLFLOW_RUN_NAME=run_name
    """

    def __init__(self) -> None:
        import mlflow

        self.tracking_uri: str = os.getenv("MLFLOW_TRACKING_URI", "")
        """The URI of the mlflow tracking server.

        An empty string, or a local file path, prefixed with file:/.
        Data is stored locally at the provided file (or ./mlruns if empty).
        """

        self.experiment_name: str = os.getenv("MLFLOW_EXPERIMENT", str(uuid4()))
        """The name of the experiment.

        If None or empty, a new experiment is created with a random UUID.
        """

        self.run_name: str = os.getenv("MLFLOW_RUN_NAME", str(uuid4()))
        """The name of the run."""

        mlflow.set_tracking_uri(self.tracking_uri)

        # activate existing or create experiment
        exp_filter_string = f"name = '{self.experiment_name}'"
        if not mlflow.search_experiments(filter_string=exp_filter_string):
            mlflow.create_experiment(name=self.experiment_name)

        self.experiment = mlflow.set_experiment(self.experiment_name)
        self.run = mlflow.start_run(run_name=self.run_name, nested=False)


@dataclass
class FeatureMapConfig:
    num_features: int = 0
    """
    Number of feature parameters to be encoded.

    Defaults to 0. Thus, no feature parameters are encoded.
    """

    basis_set: BasisSet | dict[str, BasisSet] = BasisSet.FOURIER
    """
    Basis set for feature encoding.

    Takes qadence.BasisSet.
    Give a single BasisSet to use the same for all features.
    Give a dict of (str, BasisSet) where the key is the name of the variable and the
    value is the BasisSet to use for encoding that feature.
    BasisSet.FOURIER for Fourier encoding.
    BasisSet.CHEBYSHEV for Chebyshev encoding.
    """

    reupload_scaling: ReuploadScaling | dict[str, ReuploadScaling] = ReuploadScaling.CONSTANT
    """
    Scaling for encoding the same feature on different qubits.

    Scaling used to encode the same feature on different qubits in the
    same layer of the feature maps. Takes qadence.ReuploadScaling.
    Give a single ReuploadScaling to use the same for all features.
    Give a dict of (str, ReuploadScaling) where the key is the name of the variable and the
    value is the ReuploadScaling to use for encoding that feature.
    ReuploadScaling.CONSTANT for constant scaling.
    ReuploadScaling.TOWER for linearly increasing scaling.
    ReuploadScaling.EXP for exponentially increasing scaling.
    """

    feature_range: tuple[float, float] | dict[str, tuple[float, float]] | None = None
    """
    Range of data that the input data is assumed to come from.

    Give a single tuple to use the same range for all features.
    Give a dict of (str, tuple) where the key is the name of the variable and the
    value is the feature range to use for that feature.
    """

    target_range: tuple[float, float] | dict[str, tuple[float, float]] | None = None
    """
    Range of data the data encoder assumes as natural range.

    Give a single tuple to use the same range for all features.
    Give a dict of (str, tuple) where the key is the name of the variable and the
    value is the target range to use for that feature.
    """

    multivariate_strategy: MultivariateStrategy = MultivariateStrategy.PARALLEL
    """
    The  encoding strategy in case of multi-variate function.

    Takes qadence.MultivariateStrategy.
    If PARALLEL, the features are encoded in one block of rotation gates
    with each feature given an equal number of qubits.
    If SERIES, the features are encoded sequentially, with an ansatz block
    between. PARALLEL is allowed only for DIGITAL `feature_map_strategy`.
    """

    feature_map_strategy: Strategy = Strategy.DIGITAL
    """
    Strategy for feature map.

    Accepts DIGITAL, ANALOG or RYDBERG. Defaults to DIGITAL.
    If the strategy is incompatible with the `operation` chosen, then `operation`
    gets preference and the given strategy is ignored.
    """

    param_prefix: str | None = None
    """
    String prefix to create trainable parameters in Feature Map.

    A string prefix to create trainable parameters multiplying the feature parameter
    inside the feature-encoding function. Note that currently this does not take into
    account the domain of the feature-encoding function.
    Defaults to `None` and thus, the feature map is not trainable.
    Note that this is separate from the name of the parameter.
    The user can provide a single prefix for all features, and they will be appended
    by appropriate feature name automatically.
    """

    num_repeats: int | dict[str, int] = 0
    """
    Number of feature map layers repeated in the data reuploadig step.

    If all are to be repeated the same number of times, then can give a single
    `int`. For different number of repeatitions for each feature, provide a dict
    of (str, int) where the key is the name of the variable and the value is the
    number of repeatitions for that feature.
    This amounts to the number of additional reuploads. So if `num_repeats` is N,
    the data gets uploaded N+1 times. Defaults to no repeatition.
    """

    operation: Callable[[Parameter | Basic], AnalogBlock] | Type[RX] | None = None
    """
    Type of operation.

    Choose among the analog or digital rotations or a custom
    callable function returning an AnalogBlock instance. If the type of operation is
    incompatible with the `strategy` chosen, then `operation` gets preference and
    the given strategy is ignored.
    """

    inputs: list[Basic | str] | None = None
    """
    List that indicates the order of variables of the tensors that are passed.

    Optional if a single feature is being encoded, required otherwise. Given input tensors
    `xs = torch.rand(batch_size, input_size:=2)` a QNN with `inputs=["t", "x"]` will
    assign `t, x = xs[:,0], xs[:,1]`.
    """

    def __post_init__(self) -> None:
        if self.multivariate_strategy == MultivariateStrategy.PARALLEL and self.num_features > 1:
            assert (
                self.feature_map_strategy == Strategy.DIGITAL
            ), "For `parallel` encoding of multiple features, the `feature_map_strategy` must be \
                  of `digital` type."

        if self.operation is None:
            if self.feature_map_strategy == Strategy.DIGITAL:
                self.operation = RX
            elif self.feature_map_strategy == Strategy.ANALOG:
                self.operation = AnalogRX  # type: ignore[assignment]

        else:
            if self.feature_map_strategy == Strategy.DIGITAL:
                if isinstance(self.operation, AnalogBlock):
                    logger.warning(
                        "The `operation` is of type `AnalogBlock` but the `feature_map_strategy` is\
                        `digital`. The `feature_map_strategy` will be modified and given operation\
                        will be used."
                    )

                    self.feature_map_strategy = Strategy.ANALOG

            elif self.feature_map_strategy == Strategy.ANALOG:
                if isinstance(self.operation, ParametricBlock):
                    logger.warning(
                        "The `operation` is a digital gate but the `feature_map_strategy` is\
                        `analog`. The `feature_map_strategy` will be modified and given operation\
                        will be used."
                    )

                    self.feature_map_strategy = Strategy.DIGITAL

        if self.inputs is not None:
            assert (
                len(self.inputs) == self.num_features
            ), "Inputs list must be of same size as the number of features"
        else:
            if self.num_features == 0:
                self.inputs = []
            elif self.num_features == 1:
                self.inputs = ["x"]
            else:
                raise ValueError(
                    """
                    Your QNN has more than one input. Please provide a list of inputs in the order
                    of your tensor domain. For example, if you want to pass
                    `xs = torch.rand(batch_size, input_size:=3)` to you QNN, where
                    ```
                    t = x[:,0]
                    x = x[:,1]
                    y = x[:,2]
                    ```
                    you have to specify
                    ```
                    inputs=["t", "x", "y"]
                    ```
                    You can also pass a list of sympy symbols.
                """
                )

        property_list = [
            "basis_set",
            "reupload_scaling",
            "feature_range",
            "target_range",
            "num_repeats",
        ]

        for target_field in fields(self):
            if target_field.name in property_list:
                prop = getattr(self, target_field.name)
                if isinstance(prop, dict):
                    assert set(prop.keys()) == set(
                        self.inputs
                    ), f"The keys in {target_field.name} must be the same as the inputs provided. \
                    Alternatively, provide a single value of {target_field.name} to use the same\
                    {target_field.name} for all features."
                else:
                    prop = {key: prop for key in self.inputs}
                    setattr(self, target_field.name, prop)


@dataclass
class AnsatzConfig:
    depth: int = 1
    """Number of layers of the ansatz."""

    ansatz_type: AnsatzType = AnsatzType.HEA
    """What type of ansatz.

    HEA for Hardware Efficient Ansatz.
    IIA for Identity intialized Ansatz.
    """

    ansatz_strategy: Strategy = Strategy.DIGITAL
    """Ansatz strategy.

    DIGITAL for fully digital ansatz. Required if `ansatz_type` is `iia`.
    SDAQC for analog entangling block.
    RYDBERG for fully rydberg hea ansatz.
    """

    strategy_args: dict = field(default_factory=dict)
    """
    A dictionary containing keyword arguments to the function creating the ansatz.

    Details about each below.

    For DIGITAL strategy, accepts the following:
        periodic (bool): if the qubits should be linked periodically.
            periodic=False is not supported in emu-c.
        operations (list): list of operations to cycle through in the
            digital single-qubit rotations of each layer.
            Defaults to  [RX, RY, RX] for hea and [RX, RY] for iia.
        entangler (AbstractBlock): 2-qubit entangling operation.
            Supports CNOT, CZ, CRX, CRY, CRZ, CPHASE. Controlld rotations
            will have variational parameters on the rotation angles.
            Defaults to CNOT

    For SDAQC strategy, accepts the following:
        operations (list): list of operations to cycle through in the
            digital single-qubit rotations of each layer.
            Defaults to  [RX, RY, RX] for hea and [RX, RY] for iia.
        entangler (AbstractBlock): Hamiltonian generator for the
            analog entangling layer. Time parameter is considered variational.
            Defaults to NN interaction.

    For RYDBERG strategy, accepts the following:
        addressable_detuning: whether to turn on the trainable semi-local addressing pattern
            on the detuning (n_i terms in the Hamiltonian).
            Defaults to True.
        addressable_drive: whether to turn on the trainable semi-local addressing pattern
            on the drive (sigma_i^x terms in the Hamiltonian).
            Defaults to False.
        tunable_phase: whether to have a tunable phase to get both sigma^x and sigma^y rotations
            in the drive term. If False, only a sigma^x term will be included in the drive part
            of the Hamiltonian generator.
            Defaults to False.
    """
    # The default for a dataclass can not be a mutable object without using this default_factory.

    param_prefix: str = "theta"
    """The base bame of the variational parameter."""

    def __post_init__(self) -> None:
        if self.ansatz_type == AnsatzType.IIA:
            assert (
                self.ansatz_strategy != Strategy.RYDBERG
            ), "Rydberg strategy not allowed for Identity-initialized ansatz."



from __future__ import annotations

from pulser.channels.channels import Rydberg
from pulser.channels.eom import RydbergBeam, RydbergEOM
from pulser.devices._device_datacls import VirtualDevice

from qadence.types import PI


# Idealized virtual device
def IdealDevice(
    rydberg_level: int = 60, max_abs_detuning: float = 2 * PI * 4, max_amp: float = 2 * PI * 3
) -> VirtualDevice:
    return VirtualDevice(
        name="IdealizedDevice",
        dimensions=2,
        rydberg_level=rydberg_level,
        max_atom_num=100,
        max_radial_distance=100,
        min_atom_distance=0,
        channel_objects=(
            Rydberg.Global(max_abs_detuning=max_abs_detuning, max_amp=max_amp),
            Rydberg.Local(max_targets=1000, max_abs_detuning=max_abs_detuning, max_amp=max_amp),
        ),
    )


# Device with realistic specs with local channels and custom bandwith.
def RealisticDevice(
    rydberg_level: int = 60, max_abs_detuning: float = 2 * PI * 4, max_amp: float = 2 * PI * 3
) -> VirtualDevice:
    return VirtualDevice(
        name="RealisticDevice",
        dimensions=2,
        rydberg_level=rydberg_level,
        max_atom_num=100,
        max_radial_distance=60,
        min_atom_distance=5,
        channel_objects=(
            Rydberg.Global(
                max_abs_detuning=max_abs_detuning,
                max_amp=max_amp * 0.5,
                clock_period=4,
                min_duration=16,
                max_duration=4000,
                mod_bandwidth=16,
                eom_config=RydbergEOM(
                    limiting_beam=RydbergBeam.RED,
                    max_limiting_amp=40 * 2 * PI,
                    intermediate_detuning=700 * 2 * PI,
                    mod_bandwidth=24,
                    controlled_beams=(RydbergBeam.BLUE,),
                ),
            ),
            Rydberg.Local(
                max_targets=20,
                max_abs_detuning=max_abs_detuning,
                max_amp=max_amp,
                clock_period=4,
                min_duration=16,
                max_duration=2**26,
                mod_bandwidth=16,
                eom_config=RydbergEOM(
                    limiting_beam=RydbergBeam.RED,
                    max_limiting_amp=40 * 2 * PI,
                    intermediate_detuning=700 * 2 * PI,
                    mod_bandwidth=24,
                    controlled_beams=(RydbergBeam.BLUE,),
                ),
            ),
        ),
    )



from __future__ import annotations

from collections import Counter
from functools import reduce
from typing import Dict, List, Tuple

import numpy as np
import torch
from sympy import Basic
from torch import Tensor

from qadence.backend import Backend
from qadence.backends.pyqtorch import Backend as PyQBackend
from qadence.blocks import AbstractBlock, PrimitiveBlock, chain
from qadence.circuit import QuantumCircuit
from qadence.engines.differentiable_backend import DifferentiableBackend
from qadence.noise import Noise
from qadence.operations import H, SDagger, X, Y, Z
from qadence.parameters import evaluate
from qadence.utils import Endianness


def get_qubit_indices_for_op(
    pauli_term: Tuple[AbstractBlock, Basic], op: PrimitiveBlock | None = None
) -> List[int]:
    """Get qubit indices for the given op in the Pauli term if any.

    Args:
        pauli_term: Tuple of a Pauli block and a parameter.
        op: Tuple of Primitive blocks or None.

    Returns: A list of integers representing qubit indices.
    """
    blocks = getattr(pauli_term[0], "blocks", None)
    blocks = blocks if blocks is not None else [pauli_term[0]]
    indices = [
        block.qubit_support[0] for block in blocks if (op is None) or (isinstance(block, type(op)))
    ]
    return indices


def get_counts(samples: List, support: List[int]) -> List[Counter]:
    """Marginalise the probability mass function to the support.

    Args:
        samples: List of samples against which expectation value is to be computed.
        support: A list of integers representing qubit indices.

    Returns: A List[Counter] of bit strings.
    """
    return [
        reduce(
            lambda x, y: x + y,
            [Counter({"".join([k[i] for i in support]): sample[k]}) for k, v in sample.items()],
        )
        for sample in samples
    ]


def empirical_average(samples: List, support: List[int]) -> Tensor:
    """Compute the empirical average.

    Args:
        samples: List of samples against which expectation value is to be computed.
        support: A list of integers representing qubit indices.

    Returns: A torch.Tensor of the empirical average.
    """
    PARITY = -1
    counters = get_counts(samples, support)
    n_shots = np.sum(list(counters[0].values()))
    expectations = []
    for counter in counters:
        counter_exps = []
        for bitstring, count in counter.items():
            counter_exps.append(count * PARITY ** (np.sum([int(bit) for bit in bitstring])))
        expectations.append(np.sum(counter_exps) / n_shots)
    return torch.tensor(expectations)


def pauli_z_expectation(
    pauli_decomposition: List[Tuple[AbstractBlock, Basic]],
    samples: List[Counter],
) -> Tensor:
    """Estimate total expectation value from samples by averaging all Pauli terms.

    Args:
        pauli_decomposition: A list of Pauli decomposed terms.
        samples: Samples of bit string as Counters.

    Returns: A torch.Tensor of the total expectation.
    """

    estimated_values = []
    for pauli_term in pauli_decomposition:
        support = get_qubit_indices_for_op(pauli_term)
        estim_values = empirical_average(samples=samples, support=support)
        # TODO: support for parametric observables to be tested
        estimated_values.append(estim_values * evaluate(pauli_term[1]))
    res = torch.sum(torch.stack(estimated_values), axis=0)
    return res


def rotate(circuit: QuantumCircuit, pauli_term: Tuple[AbstractBlock, Basic]) -> QuantumCircuit:
    """Rotate circuit to measurement basis and return the qubit support.

    Args:
        circuit: The circuit that is executed.
        pauli_term: Tuple of a Pauli term and a parameter.

    Returns: Rotated QuantumCircuit.
    """

    rotations = []

    for op, gate in [(X(0), Z), (Y(0), SDagger)]:
        qubit_indices = get_qubit_indices_for_op(pauli_term, op=op)
        for index in qubit_indices:
            rotations.append(gate(index) * H(index))
    rotated_block = chain(circuit.block, *rotations)
    return QuantumCircuit(circuit.register, rotated_block)


def iterate_pauli_decomposition(
    circuit: QuantumCircuit,
    param_values: Dict[str, Tensor],
    pauli_decomposition: List[Tuple[AbstractBlock, Basic]],
    n_shots: int,
    state: Tensor | None = None,
    backend: Backend | DifferentiableBackend = PyQBackend(),
    noise: Noise | None = None,
    endianness: Endianness = Endianness.BIG,
) -> Tensor:
    """Estimate total expectation value by averaging all Pauli terms.

    Args:
        circuit: The circuit that is executed.
        param_values: Parameters of the circuit.
        pauli_decomposition: A list of Pauli decomposed terms.
        n_shots: Number of shots to sample.
        state: Initial state.
        backend: A backend for circuit execution.
        noise: A noise model to use.
        endianness: Endianness of the resulting bit strings.

    Returns: A torch.Tensor of bit strings n_shots x n_qubits.
    """

    estimated_values = []

    for pauli_term in pauli_decomposition:
        if pauli_term[0].is_identity:
            estimated_values.append(evaluate(pauli_term[1], as_torch=True))
        else:
            # Get the full qubit support for the Pauli term.
            # Note: duplicates must be kept here to allow for
            # observables chaining multiple operations on the same qubit
            # such as `b = chain(Z(0), Z(0))`
            support = get_qubit_indices_for_op(pauli_term)
            # Rotate the circuit according to the given observable term.
            rotated_circuit = rotate(circuit=circuit, pauli_term=pauli_term)
            # Use the low-level backend API to avoid embedding of parameters
            # already performed at the higher QuantumModel level.
            # Therefore, parameters passed here have already been embedded.
            conv_circ = backend.circuit(rotated_circuit)
            samples = backend.sample(
                circuit=conv_circ,
                param_values=param_values,
                n_shots=n_shots,
                state=state,
                noise=noise,
                endianness=endianness,
            )
            estim_values = empirical_average(samples=samples, support=support)
            # TODO: support for parametric observables to be tested
            estimated_values.append(estim_values * evaluate(pauli_term[1]))
    res = torch.sum(torch.stack(estimated_values), axis=0)
    # Allow for automatic differentiation.
    res.requires_grad = True
    return res



from __future__ import annotations

import os
import re
from logging import getLogger
from pathlib import Path
from typing import Any

import torch
from nevergrad.optimization.base import Optimizer as NGOptimizer
from torch.nn import Module
from torch.optim import Optimizer

logger = getLogger(__name__)


def get_latest_checkpoint_name(folder: Path, type: str, device: str | torch.device = "cpu") -> Path:
    file = Path("")
    files = [f for f in os.listdir(folder) if f.endswith(".pt") and type in f]
    if len(files) == 0:
        logger.error(f"Directory {folder} does not contain any {type} checkpoints.")
    if len(files) == 1:
        file = Path(files[0])
    else:
        device = str(device).split(":")[0]
        pattern = re.compile(f".*_(\d+)_device_{device}.pt$")
        legacy_pattern = re.compile(".*_(\d+).pt$")
        max_index = -1
        for f in files:
            legacy_match = legacy_pattern.search(f)
            match = pattern.search(f)
            if match or legacy_match:
                if legacy_match:
                    logger.warn(f"Found checkpoint(s) in legacy format: {f}.")
                    match = legacy_match
                index_str = match.group(1).replace("_", "")  # type: ignore [union-attr]
                index = int(index_str)
                if index > max_index:
                    max_index = index
                    file = Path(f)
    return Path(file)


def load_checkpoint(
    folder: Path,
    model: Module,
    optimizer: Optimizer | NGOptimizer,
    model_ckpt_name: str | Path = "",
    opt_ckpt_name: str | Path = "",
    device: str | torch.device = "cpu",
) -> tuple[Module, Optimizer | NGOptimizer, int]:
    if isinstance(folder, str):
        folder = Path(folder)
    if not folder.exists():
        folder.mkdir(parents=True)
        return model, optimizer, 0
    model, iter = load_model(folder, model, model_ckpt_name, device)
    optimizer = load_optimizer(folder, optimizer, opt_ckpt_name, device)
    return model, optimizer, iter


def write_checkpoint(
    folder: Path,
    model: Module,
    optimizer: Optimizer | NGOptimizer,
    iteration: int | str,
) -> None:
    from qadence import QuantumModel

    from .models import QNN

    device = None
    try:
        # We extract the device from the pyqtorch native circuit
        device = str(model.device).split(":")[0]  # in case of using several CUDA devices
    except Exception as e:
        msg = (
            f"Unable to identify in which device the QuantumModel is stored due to {e}."
            "Setting device to None"
        )
        logger.warning(msg)

    iteration_substring = f"{iteration:03n}" if isinstance(iteration, int) else iteration
    model_checkpoint_name: str = (
        f"model_{type(model).__name__}_ckpt_"
        + f"{iteration_substring}"
        + f"_device_{device}"
        + ".pt"
    )
    opt_checkpoint_name: str = (
        f"opt_{type(optimizer).__name__}_ckpt_"
        + f"{iteration_substring}"
        + f"_device_{device}"
        + ".pt"
    )
    try:
        d = (
            model._to_dict(save_params=True)
            if isinstance(model, (QNN, QuantumModel))
            else model.state_dict()
        )
        torch.save((iteration, d), folder / model_checkpoint_name)
        logger.info(f"Writing {type(model).__name__} checkpoint {model_checkpoint_name}")
    except Exception as e:
        logger.exception(e)
    try:
        if isinstance(optimizer, Optimizer):
            torch.save(
                (iteration, type(optimizer), optimizer.state_dict()), folder / opt_checkpoint_name
            )
        elif isinstance(optimizer, NGOptimizer):
            optimizer.dump(folder / opt_checkpoint_name)
        logger.info(f"Writing {type(optimizer).__name__} to checkpoint {opt_checkpoint_name}")
    except Exception as e:
        logger.exception(e)


def load_model(
    folder: Path,
    model: Module,
    model_ckpt_name: str | Path = "",
    device: str | torch.device = "cpu",
    *args: Any,
    **kwargs: Any,
) -> tuple[Module, int]:
    from qadence import QNN, QuantumModel

    iteration = 0
    if model_ckpt_name == "":
        model_ckpt_name = get_latest_checkpoint_name(folder, "model", device)

    try:
        iteration, model_dict = torch.load(folder / model_ckpt_name, *args, **kwargs)
        if isinstance(model, (QuantumModel, QNN)):
            model._from_dict(model_dict, as_torch=True)
        elif isinstance(model, Module):
            model.load_state_dict(model_dict, strict=True)
        # Load model to a specific gpu device if specified
        pattern = re.compile("cuda:\d+$")
        if pattern.search(str(device)):
            model.to(device)

    except Exception as e:
        msg = f"Unable to load state dict due to {e}.\
               No corresponding pre-trained model found. Returning the un-trained model."
        logger.warning(msg)
    return model, iteration


def load_optimizer(
    folder: Path,
    optimizer: Optimizer | NGOptimizer,
    opt_ckpt_name: str | Path = "",
    device: str | torch.device = "cpu",
) -> Optimizer | NGOptimizer:
    if opt_ckpt_name == "":
        opt_ckpt_name = get_latest_checkpoint_name(folder, "opt", device)
    if os.path.isfile(folder / opt_ckpt_name):
        if isinstance(optimizer, Optimizer):
            (_, OptType, optimizer_state) = torch.load(folder / opt_ckpt_name)
            if isinstance(optimizer, OptType):
                optimizer.load_state_dict(optimizer_state)

        elif isinstance(optimizer, NGOptimizer):
            loaded_optimizer = NGOptimizer.load(folder / opt_ckpt_name)
            if loaded_optimizer.name == optimizer.name:
                optimizer = loaded_optimizer
        else:
            raise NotImplementedError
    return optimizer



from __future__ import annotations

from dataclasses import dataclass, field
from logging import getLogger
from typing import Callable

from qadence.backend import BackendConfiguration
from qadence.transpile import digitalize, fill_identities

logger = getLogger(__name__)

default_passes: list[Callable] = [fill_identities, digitalize]


@dataclass
class Configuration(BackendConfiguration):
    # FIXME: currently not used
    cloud_credentials: dict = field(default_factory=dict)
    """Credentials for connecting to the cloud.

    Execution on the QPUs available on Amazon Braket.
    """
